/*
 * Author: wangjian
 * 生成ServiceImpl
 */
package com.tofly.xrtymis.custom.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.tofly.api.UserClent;

import com.tofly.common.core.entity.ResultRespone;
import com.tofly.pojo.User;
import com.tofly.pojo.comm.Audit;
import com.tofly.pojo.plan.Checkhistory;
import com.tofly.pojo.plan.vo.CheckhistoryVo;
import com.tofly.xrtymis.custom.commons.AuditSateEnum;
import com.tofly.xrtymis.custom.commons.AuditTypeEnum;
import com.tofly.xrtymis.custom.service.CheckhistoryService;
import com.tofly.xrtymis.custom.service.IAuditService;
import com.tofly.xrtymis.mapper.CheckhistoryMapper;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.RequestBody;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;


/**
 * 审核记录历史表
 *
 * @author admin
 * @date Tue Mar 30 00:00:00 CST 2021
 */
@Service
@Slf4j
public class CheckhistoryServiceImpl extends ServiceImpl<CheckhistoryMapper, Checkhistory> implements CheckhistoryService {

    @Autowired
    private UserClent userClent;

    @Autowired
    private IAuditService iAuditService;

    @Override
    public List<CheckhistoryVo> viewConversion(List<Checkhistory> infoList) {
        List<CheckhistoryVo> infoVoList = new ArrayList<>();
        infoList.forEach(info -> {
            CheckhistoryVo baseInfo = JSON.parseObject(JSON.toJSONString(info), CheckhistoryVo.class);
            // 转换审核类型/审核意见
            baseInfo.setType(AuditTypeEnum.getCodeForValue(info.getType()));
            baseInfo.setAuditResult(AuditSateEnum.getCodeForValue(info.getAuditResult()));
            Long auditUser = info.getAuditUser();
            if (null != auditUser && 0 != auditUser) {
                User userAudit = userClent.getById(auditUser).getResult();
                if (null != userAudit) {
                    baseInfo.setAuditUserName(userAudit.getRealName());
                }

            }
            infoVoList.add(baseInfo);
        });
        return infoVoList.stream().sorted(Comparator.comparing(CheckhistoryVo::getAuditTime).reversed()).collect(Collectors.toList());
    }

    @Override
    public boolean recordHis(String type, Long checkId, String auditResult, String auditNotes, Long auditUser, Date auditTime) {
        Checkhistory info = new Checkhistory();
        info.setType(type);
        info.setCheckId(checkId);
        info.setAuditResult(auditResult);
        info.setAuditNotes(auditNotes);
        info.setAuditUser(auditUser);
        info.setAuditTime(auditTime);
        return this.save(info);
    }

    @Override
    public boolean delete(String type, Long planId) {
        Audit audit = new Audit();
        audit.setAuditType(type);
        audit.setAuditPlanId(planId);
        log.info("删除对应审核信息 类型:{},计划id{}", type, planId);
        boolean b = iAuditService.remove(new QueryWrapper<Audit>().setEntity(audit));
        return b;
    }

    @Override
    public ResultRespone<List<Audit>> getAuditInfo(Audit audit) {
        ResultRespone<List<Audit>> rest = new ResultRespone<>();
        List<Audit> list = iAuditService.list(new QueryWrapper<>(audit));
        rest.setResult(list);
        return rest;
    }

    @Override
    public boolean update(Audit audit) {
        boolean b = iAuditService.updateById(audit);
        return b;
    }

    @Override
    public boolean saveAuditHisInfo(Checkhistory checkhistory) {
        boolean save = this.recordHis(checkhistory.getType(), checkhistory.getCheckId(), checkhistory.getAuditResult(), checkhistory.getAuditNotes(), checkhistory.getAuditUser(), checkhistory.getAuditTime());
        return save;
    }
    @Override
    public boolean save(String json){
        List<Audit> auditList = JSONArray.parseArray(json, Audit.class);
        auditList.forEach(info -> iAuditService.save(info));
        return true;
    }
    @Override
    public ResultRespone<Object> getAuditHisInfo( Checkhistory checkhistory) {
        ResultRespone<Object> rest = new ResultRespone<>();
        List<Checkhistory> list = this.list(new QueryWrapper<>(checkhistory));
        rest.setResult(list);
        return rest;
    }


}
